/**********************************************************************
** Copyright (C) 2000-2002 Trolltech AS.  All rights reserved.
**
** This file is part of the Qtopia Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/

#include <qtopia/qpeapplication.h>
#include <qtopia/config.h>
#include <qtopia/network.h>
#ifdef QWS
#include <qtopia/qcopenvelope_qws.h>
#endif
#include <qtopia/global.h>
#include <qtopia/process.h>
#include "../proxiespage.h"

#include <qbuttongroup.h>
#include <qvbox.h>
#include <qradiobutton.h>
#include <qlineedit.h>
#include <qcombobox.h>
#include <qcheckbox.h>
#include <qlabel.h>
#include <qvalidator.h>
#include <qtextstream.h>
#include <qfile.h>
#include <qtabwidget.h>
#include <qbuffer.h>
#include <qspinbox.h>
#include <qmessagebox.h>
#include <qregexp.h>
#include <qlayout.h>

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/vfs.h>
#include <mntent.h>
#include <unistd.h>
#include <sys/types.h>

#include "pan.h"
#include "panstate.h"

#define USE_SCHEMES

#ifdef USE_SCHEMES
static QString toSchemeId(const QString& s)
{
	QString r;
	for (int i=0; i<(int)s.length(); i++) {
		char c = s[i].lower().latin1();
		if ( c>='a' && c<='z' || i>0 && c>='0' && c<='9' )
			r += s[i];
		//else
		//r += "_U" + QString::number(s[i].unicode()) + "_";
	}
	return r;
}
#endif

class SchemeChanger : public QObject {
	Q_OBJECT
public:
	SchemeChanger(QObject* parent) :
	QObject(parent)
	{
	}

	static void changePanScheme(const QString& pan_or_scheme)
	{
#ifdef USE_SCHEMES	
		Config c("Network");
		c.setGroup("Pan");
		QString s = toSchemeId(pan_or_scheme);
		c.writeEntry("Scheme",s);
		system("cardctl scheme " + s);
#else	
		QStringList srv = Network::choices();
		for (QStringList::ConstIterator it=srv.begin(); it!=srv.end(); ++it) {
			QString t = Network::serviceType(*it);
			if ( t == "pan" ) {
				Config cfg(*it,Config::File);
				cfg.setGroup("Info");
				QString n = cfg.readEntry("Name");
				if ( n == pan_or_scheme ) {
					Pan::writeNetworkOpts( cfg );
				}
			}
		}	
#endif	
#ifndef QT_NO_COP
		QCopEnvelope e("QPE/Network","choicesChanged()");
#endif
	}

public slots:
	void changePanSchemeSlot(const QString& pan_or_scheme)
	{
		changePanScheme(pan_or_scheme);
	}
};

class WepKeyValidator : public QValidator {
public:
	WepKeyValidator(QWidget* parent) :
	  QValidator(parent)
	  {
	  }

	  State validate( QString& key, int& curs ) const
	  {
		  QString k;
		  int hexes=0;
		  int ncurs=0;
		  for (int i=0; i<(int)key.length(); i++) {
			  char c=key[i].lower().latin1();
			  if ( c>='0' && c<='9' || c>='a' && c<='z' ) {
				  if ( hexes == 2 ) {
					  hexes = 0;
					  k += ':';
					  if ( i<curs ) ncurs++;
				  }
				  k += c;
				  hexes++;
				  if ( i<curs ) ncurs++;
			  } else if ( c == ':' && hexes==2 ) {
				  hexes = 0;
				  k += c;
				  if ( i<curs ) ncurs++;
			  }
		  }
		  key = k;
		  curs = ncurs;
		  return Acceptable;
	  }

	  //void fixup( QString& key ) const
	  //{
	  //}
};

Pan::Pan(Config& cfg, QWidget* parent) :
PanBase(parent,0,TRUE),
config(cfg)
{
	ipGroup = new QButtonGroup( this );
	ipGroup->hide();
	int id = ipGroup->insert( manualIP );
	ipGroup->insert( autoIP );
	autoIP->setEnabled(false);	// not yet implemented...
	connect( ipGroup, SIGNAL(clicked(int)), this, SLOT(ipSelect(int)) );
	ipGroup->setButton(id);

	QBoxLayout *proxyLayout  = new QVBoxLayout( tab_2 ); 
	proxies = new ProxiesPage( tab_2 );
	proxyLayout->addWidget( proxies );

	tabs->removePage(wireless);
	
	readConfig();
}

void Pan::btSearchPressed()
{
    system("/usr/bin/sdptool browse 2>/tmp/napsearch.err >/tmp/napsearch.txt");

    FILE *f = fopen("/tmp/napsearch.txt", "rb");
    if (f) {
	char buf[256];
	while (!feof(f)) {
	    fgets(buf, 32, f);
	    if (strncmp(buf, "Browsing", 8) == 0) {
		char mac[32];
		strncpy(mac, buf+9, 17);
		mac[17] = 0;
		if ((mac[2]==':') && (mac[5]==':') && (mac[8]==':') && (mac[11]==':') && (mac[14]==':')) {
		    btAddress->setText(mac);
		    fclose(f);
		    unlink("/tmp/napsearch.txt");
		    unlink("/tmp/napsearch.err");
		    return;
		} else goto got_err;
	    }
	}
	fclose(f);
    }
  got_err:

    unlink("/tmp/napsearch.txt");
    unlink("/tmp/napsearch.err");
    QMessageBox::information(this, tr("NAP Search"), tr("Can't get NAP info, try run search manually."));  
}

void Pan::ipSelect( int id )
{
	tabs->setTabEnabled( tcpip, id==0 );
}

void Pan::readConfig()
{
	int id = config.readEntry("DHCP","n") != "n";
	ipGroup->setButton( id );
	ipSelect( id );

	config.setGroup("Info");
	acname->setText(config.readEntry("Name", "PAN"));

	config.setGroup("Properties");
	ip->setText(config.readEntry("IPADDR"));
	subnet->setText( config.readEntry("NETMASK", "255.255.255.0"));
	broadcast->setText( config.readEntry("BROADCAST", ""));
	gateway->setText( config.readEntry("GATEWAY"));
	dns1->setText( config.readEntry("DNS_1"));
	dns2->setText( config.readEntry("DNS_2"));

	config.setGroup("Proxy");
	proxies->readConfig( config );

	// bluetooth device address
	config.setGroup("Bluetooth");
	btAddress->setText(config.readEntry("btAddress",""));

}

bool Pan::writeConfig()
{
	QString nm = acname->text();
	config.setGroup("Info");

	// make sure we don't duplicate the name
	QStringList srv = Network::choices();
	for (QStringList::ConstIterator it=srv.begin(); it!=srv.end(); ++it) 
	{
		QString t = Network::serviceType(*it);
		if ( t == "pan" ) 
		{
			Config cfg(*it,Config::File);
			cfg.setGroup("Info");
			QString n = cfg.readEntry("Name");
			if ( (n == nm
/*
#ifdef USE_SCHEMES
				|| toSchemeId(n) == toSchemeId(nm)
#endif
*/
				) && (nm != config.readEntry("Name"))) // did not change the name
			{
				QMessageBox::warning(0, tr("PAN Setup"), 
					tr( "This name already\nexists. Please choose a\ndifferent name.") );
				return FALSE;
			}
		}
	}

#ifdef USE_SCHEMES    
	QString scheme;
#endif
	{
		config.setGroup("Info");
		config.writeEntry("Name", nm);
#ifdef USE_SCHEMES    
		scheme = toSchemeId(nm);
		if ( scheme.isEmpty() )
			scheme = "*"; // The default config
#endif    
		config.setGroup("Properties");

		// check requried
		if (manualIP->isChecked())
		{
			if (ip->text() == "")
			{
				QMessageBox::warning( this, tr("PAN Setup"), tr("IP address is requried.") );
				tabs->showPage(tcpip);
				return FALSE;
			}
		}

		config.writeEntry("DHCP",autoIP->isChecked() ? "y" : "n");
		config.writeEntry("IPADDR",ip->text());
		config.writeEntry("NETMASK",subnet->text());
		config.writeEntry("BROADCAST",broadcast->text());
		config.writeEntry("GATEWAY",gateway->text());
		config.writeEntry("DNS_1",dns1->text());
		config.writeEntry("DNS_2",dns2->text());

		config.setGroup("Proxy");
		proxies->writeConfig( config );

		// bluetooth device address
		if (btAddress->text() == "")
		{
			QMessageBox::warning( this, tr("PAN Setup"), tr("NAP device address is requried.") );
			tabs->showPage(bluetooth);
			return FALSE;
		}

		config.setGroup("Bluetooth");
		config.writeEntry("btAddress", btAddress->text());
	}

#ifdef USE_SCHEMES    
	return writeNetworkOpts( config, scheme );
#else
	return TRUE;
#endif    
}

bool Pan::writeNetworkOpts( Config &config, QString scheme )
{
	// PCMCIA card will have "CardType = network" entry
	config.setGroup("Properties");
	QString cardType = config.readEntry("CardType");

	if (cardType == "network") // No tr
		return writePcmciaNetworkOpts(config, scheme);
	else if (cardType == "pan") // No tr
		return writePanNetworkOpts(config, scheme);
	else
		return writeBuiltinNetworkOpts(config, scheme);
}

bool Pan::writePanNetworkOpts( Config& /*config*/, QString /*scheme*/ )
{
	return TRUE;
}

bool Pan::writeBuiltinNetworkOpts( Config& /*config*/, QString /*scheme*/ )
{
	// Since no standard exists for configuring networks, this
	// function must be implemented by the system integrator.
	//
	// The procedure is:
	//  1. write the network configuration file to a tmp file.
	//  2. stop the network.
	//  3. move the tmp file to the network config file.
	//  4. start the network.

	return TRUE;
}

bool Pan::writePcmciaNetworkOpts( Config &config, QString scheme )
{
	QString prev = "/etc/pcmcia/network.opts";
	QFile prevFile(prev);
	if ( !prevFile.open( IO_ReadOnly ) )
		return FALSE;

	QString tmp = prev + "-qpe-new";
	QFile tmpFile(tmp);
	if ( !tmpFile.open( IO_WriteOnly ) )
		return FALSE;

	bool retval = TRUE;

	QTextStream in( &prevFile );
	QTextStream out( &tmpFile );

	config.setGroup("Info");
	QString nm = config.readEntry( "Name" );

	config.setGroup("Properties");

	//For DHCP to work, we have to remove the TCP/IP fields
	bool dhcp = config.readEntry("DHCP","n") != "n";

	QString line;
	bool found=FALSE;
	bool done=FALSE;
	while ( !in.atEnd() ) {
		QString line = in.readLine();
		QString wline = line.simplifyWhiteSpace();
		if ( !done ) {
			if ( found ) {
				if ( wline == ";;" ) {
					done = TRUE;
				} else {
					int eq=wline.find("=");
					QString k,s,v;
					if ( eq > 0 ) {
						k = wline.left(eq);
						s = "=";
					} else if ( wline.left(8) == "start_fn" ) {
						k = wline.left(8);
						s = " () ";
					} else if ( wline.left(7) == "stop_fn" ) {
						k = wline.left(7);
						s = " () ";
					}
					if ( !k.isNull() ) {
						QString v = config.readEntry(k);
						if ( eq > 0 ) {
							if ( dhcp && k.left(4) != "DHCP" )
								v = "";
							else
								v = Global::shellQuote(v);
						} else {
							if ( v.isEmpty() )
								v = "{ return; }"; // No tr
						}
						line = "    " + k + s + v;
					}
				}
			} else {
				if ( wline.left(scheme.length()+7) == scheme + ",*,*,*)" ) {
					found=TRUE;
				} else if ( wline == "esac" || wline == "*,*,*,*)" ) {
					// end - add new entry
					// Not all fields have a GUI, but all are supported
					// in the network configuration files.
					static const char* txtfields[] = {
						"IF_PORT", "DHCP_HOSTNAME", "NETWORK",
							"DOMAIN", "SEARCH", "MOUNTS",
							"MTU", "NO_CHECK", "NO_FUSER",
							"IPADDR", "NETMASK", "BROADCAST", "GATEWAY",
							"DNS_1", "DNS_2", "DNS_3", 0
					};
					out << scheme << ",*,*,*)" << "\n"
						<< "    INFO=" << Global::shellQuote(nm) << "\n"
						<< "    BOOTP=" << config.readEntry("BOOTP","n") << "\n"
						<< "    DHCP=" << config.readEntry("DHCP","n") << "\n"
						<< "    start_fn () " << config.readEntry("start_fn","{ return; }") << "\n"
						<< "    stop_fn () " << config.readEntry("stop_fn","{ return; }") << "\n"
						;
					const char** f = txtfields;
					while (*f) {
						out << "    " << *f << "=" 
							<< Global::shellQuote(dhcp ? QString("") : config.readEntry(*f,""))
							<< "\n";
						++f;
					}
					out << "    ;;\n";
					done = TRUE;
				}
			}
		}
		out << line << "\n";
	}

	prevFile.close();
	tmpFile.close();

	if ( system( "mv " + tmp + " " + prev ) )
		retval = FALSE;

#ifdef USE_SCHEMES
	if ( retval )
		SchemeChanger::changePanScheme(scheme);
#endif

	//system("cardctl resume");
	//system("cardctl insert");

	return retval;
}

void Pan::accept()
{
	if ( writeConfig() )
		QDialog::accept();
}

static QStringList findPans(const QString& skip=QString::null)
{
	QStringList srv = Network::choices();
	QStringList pans;
	for (QStringList::ConstIterator it=srv.begin(); it!=srv.end(); ++it) {
		QString t = Network::serviceType(*it);
		if ( t == "pan" ) {
			Config cfg(*it,Config::File);
			cfg.setGroup("Info");
			QString n = cfg.readEntry("Name");
			if ( n != skip )
				pans.append(n);
		}
	}
	return pans;
}

static QString findScheme(const QStringList& pans)
{
	QString scheme;
	if ( pans.count() > 0 ) {
		{
			Config c("Network");
			c.setGroup("Pan");
			scheme = c.readEntry("Scheme");
		}
		bool found=FALSE;
		for (QStringList::ConstIterator it=pans.begin(); !found && it!=pans.end(); ++it) {
			if ( toSchemeId(*it) == scheme )
				found = TRUE;
		}
		if ( !found )
			SchemeChanger::changePanScheme( pans[0] );
	} else {
		{
			Config c("Network");
			c.setGroup("Pan");
			c.writeEntry("Scheme","default"); // No tr
		}
		SchemeChanger::changePanScheme( "default" ); // No tr
	}
	return scheme;
}


PanImpl::PanImpl()
{
}

PanImpl::~PanImpl()
{
}

QRESULT PanImpl::queryInterface( const QUuid &uuid, QUnknownInterface **iface )
{
	*iface = 0;
	if ( uuid == IID_QUnknown )
		*iface = this;
	else if ( uuid == IID_Network )
		*iface = this;
	else
		return QS_FALSE;

	(*iface)->addRef();
	return QS_OK;
}

Q_EXPORT_INTERFACE()
{
	Q_CREATE_INSTANCE( PanImpl )
}

bool PanImpl::doProperties( QWidget *parent, Config& cfg )
{
	Pan dialog(cfg,parent);
	dialog.showMaximized();
	return dialog.exec();
}

bool PanImpl::create( Config& cfg )
{
	QStringList pans = findPans();
	QString nm = cfg.readEntry("Name");
	int n=1;
	while ( pans.contains(nm) ) {
		nm = "PAN " + QString::number(++n);
	}
	cfg.writeEntry("Name",nm);
	return TRUE;
}

bool PanImpl::remove( Config& cfg )
{
#ifdef USE_SCHEMES
	// Switch to another
	cfg.setGroup("Info");
	QStringList pans = findPans(cfg.readEntry("Name"));
	findScheme(pans);
#endif
	return TRUE;
}

int PanImpl::exec(const QString& arg, QTextStream& log)
{
	log << arg << endl;
	int rc = system((!isRoot ? "sudo " : "") + arg);
	if (rc == 0)
		log << "success!" << endl;
	else
		log << "failed!" << endl;

	return rc;
}

void PanImpl::progress(const QString& p, int flag)
{
#ifndef QT_NO_COP
	QCopEnvelope e("QPE/Network", "progress(QString, int)");
	e << p << flag;
#endif
}

bool PanImpl::start( Config& cfg )
{
	isRoot = (geteuid() == 0);

	// start PAN
	QFile tmp("/tmp/pan.log");
	if (!tmp.open(IO_WriteOnly | IO_Append))
		return false;

	QTextStream log( &tmp );
	log << "Starting PAN..." << endl;

	progress(QT_TR_NOOP("Initializing"), 2);
	if (exec("modprobe bnep", log))
		return false;

	cfg.setGroup("Bluetooth");
	QString btAddress = cfg.readEntry("btAddress");

	// connect to the NAP
	progress(QT_TR_NOOP("Connecting"), 5);
	QString command = "pand --role PANU --service NAP --connect %1 --nodetach";
	if (exec(command.arg(btAddress), log))
		return false;

	cfg.setGroup("Properties");

	// set up IP
	progress(QT_TR_NOOP("Configuring network"), 0);
	command = "ifconfig bnep0 %1 netmask %2";
	QString ip = cfg.readEntry("IPADDR");
	QString mask = cfg.readEntry("NETMASK","255.255.255.0");

	QString broadcast = cfg.readEntry("BROADCAST","");
	if (broadcast != "")
		command += " broadcast %3";

	if (exec(command.arg(ip).arg(mask).arg(broadcast), log))
		return false;

	system("sleep 1");

	// add route
	QString gw = cfg.readEntry("GATEWAY");
	if (gw != "")
	{
		command = "route add default gw %1";
		exec(command.arg(gw), log);	// ignore result
	}

	// add DNS
	QString dns1 =  cfg.readEntry("DNS_1");
	QString dns2 = cfg.readEntry("DNS_2");

	exec("cp -f /etc/resolv.conf /etc/resolv.conf.pan", log);

	if (dns1 != "" || dns2 != "")
	{
		log << "Adding DNS entries..." << endl;

		QFile resolve("/etc/resolv.conf");
		if (!resolve.open(IO_WriteOnly | IO_Truncate))
			return false;

		QTextStream out(&resolve);
		
		if (dns1 != "")
		{
			out << "nameserver " << dns1 << endl;
			log << "nameserver " << dns1 << endl;
		}

		if (dns2 != "")
		{
			out << "nameserver " << dns2 << endl;
			log << "nameserver " << dns2 << endl;
		}

		if (dns1 == "" && dns2 == "")
			out << "nameserver " << "127.0.0.1" << endl;

		log << "success!" << endl;
		resolve.close();
	}

	progress(QT_TR_NOOP("Connected"), 5);
	tmp.close();

	return true;
}

bool PanImpl::stop( Config& cfg )
{
	QFile tmp("/tmp/pan.log");
	if (!tmp.open(IO_WriteOnly | IO_Append))
		return false;

	QTextStream log( &tmp );

	log << "Stopping PAN..." << endl;
	exec("pand -K", log);

	QString dns1 =  cfg.readEntry("DNS_1");
	QString dns2 = cfg.readEntry("DNS_2");

	// restore config if needed
	exec("mv -f /etc/resolv.conf.pan /etc/resolv.conf", log);

	return !isActive(cfg);
}

bool PanImpl::isActive( Config& /*cfg*/ ) const
{
	QString result;
	Process p(QStringList() << "ifconfig");
	p.exec("", result);

	return result.contains("bnep");
}

bool PanImpl::isAvailable( Config& /*cfg*/ ) const
{
	QString result;
	Process p(QStringList() << "hciconfig");
	p.exec("", result);

	return result != "";
}

class PanStateImpl : public PanState 
{
	Q_OBJECT
public:
	PanStateImpl(QWidget* parent, const QString &dev, const QString &bt) : 
		PanState(parent), device(dev), btAddress(bt)
	{
		QStringList pans = findPans();
		QString scheme = findScheme(pans);
		if ( pans.count() > 1 ) {
			services->insertStringList(pans);
			services->setCurrentItem(pans.findIndex(scheme));
			SchemeChanger* sc = new SchemeChanger(this);
			QObject::connect(services,SIGNAL(activated(const QString&)),
				sc,SLOT(changePanSchemeSlot(const QString&)));
		} else {
			services->hide();
			services_label->hide();
		}
		if ( !device[device.length()-1].isDigit() )
			device += "0";
		tid = startTimer(1000);
		updateDisplay();
	}

	void timerEvent(QTimerEvent*)
	{
		updateDisplay();
		if ( ipaddress->text()[0] != 'e' ) 
		{
			// slow down
			killTimer(tid);
			tid = startTimer(5000);
		}
	}

	bool ok() const
	{
		return !ipaddress->text().isEmpty();
	}

private slots:
	void updateDisplay()
	{
		QString ipaddr;
		Process p(QStringList() << "ifconfig" << device);
		if ( p.exec( "", ipaddr ) ) 
		{
			ipaddr.replace(QRegExp(".*inet addr:"),"");
			ipaddr.replace(QRegExp(" .*"),"");
			ipaddress->setText(ipaddr);
		}
		else 
		{
			ipaddress->setText("");
		}

		btaddress->setText(btAddress);
	}

private:
	int tid;
	QString device;
	QString btAddress;
};

QWidget* PanImpl::addStateWidget( QWidget* parent, Config& cfg ) const
{
	if ( isActive(cfg) ) 
	{
		cfg.setGroup("Properties");
		QString dev = cfg.readEntry("Device");

		cfg.setGroup("Bluetooth");
		QString btAddress = cfg.readEntry("btAddress");
		PanStateImpl* state = new PanStateImpl(parent, dev, btAddress);

		if ( state->ok() )
			return state;

		delete state;
	}
	return 0;
}


#include "pan.moc"
